/********************************************************************************
 * Copyright (C) 2014 by Lorenzo Porta (Vindex17) <vindex17@outlook.it>         *
 *                                                                              *
 * This program is free software; you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as               *
 * published by the Free Software Foundation; either version 2 of               *
 * the License or (at your option) version 3 or any later version               *
 * accepted by the membership of KDE e.V. (or its successor approved            *
 * by the membership of KDE e.V.), which shall act as a proxy                   *
 * defined in Section 14 of version 3 of the license.                           *
 *                                                                              *
 * This program is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of               *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
 * GNU General Public License for more details.                                 *
 *                                                                              *
 * You should have received a copy of the GNU General Public License            *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                              *
 ********************************************************************************/

#include "kfoldersync.h"
#include "kfoldersync.moc"

#include <KAboutData>
#include <KAction>
#include <KActionCollection>
#include <KUniqueApplication>
#include <KDiskFreeSpaceInfo>
#include <KInputDialog>
#include <KIO/CopyJob>
#include <KIO/DeleteJob>
#include <KIO/JobUiDelegate>
#include <KLocale>
#include <KMessageBox>
#include <KStandardAction>
#include <KStandardDirs>
#include <KStatusBar>
#include <QThread>

#define ONE_HUNDRED 100
#define ONE_THOUSAND 1000
#define DEFAULT_SECONDS_BETWEEN_ATTEMPTS 30

KFolderSync::KFolderSync(): tray_obj(this), background_timer(this)
{
    // Tell the KXmlGuiWindow that this is indeed the main widget
    QWidget* w = new QWidget(this);
    setCentralWidget(w);
    kfoldersync_ui.setupUi(w);

    // Add a requesterwidget
    add_req_widget();
    req_widget_list.first()->set_button_enabled(false);

    // Setup actions and details
    setup_actions();
    background_timer.setSingleShot(true);

    // Setup tray object
    tray_obj.setIconByName("kfoldersync");

    // Setup UI connections
    connect(kfoldersync_ui.addButton, SIGNAL(pressed()), SLOT(add_req_widget()));
    connect(kfoldersync_ui.sync_radioButton, SIGNAL(clicked(bool)), SLOT(set_sync_enum(bool)));
    connect(kfoldersync_ui.bilateralSync_radioButton, SIGNAL(clicked(bool)), SLOT(set_bilateral_sync_enum(bool)));
    connect(kfoldersync_ui.diffBackup_radioButton, SIGNAL(clicked(bool)), SLOT(set_diff_backup_enum(bool)));
    connect(kfoldersync_ui.incrBackup_radioButton, SIGNAL(clicked(bool)), SLOT(set_incr_backup_enum(bool)));
    connect(kfoldersync_ui.bilateralIncrBackup_radioButton, SIGNAL(clicked(bool)), SLOT(set_bilateral_incr_backup_enum(bool)));
    connect(kfoldersync_ui.fullBackup_radioButton, SIGNAL(clicked(bool)), SLOT(set_full_backup_enum(bool)));
    connect(kfoldersync_ui.log_checkBox, SIGNAL(clicked(bool)), SLOT(init_logger(bool)));
    connect(kfoldersync_ui.trash_checkBox, SIGNAL(clicked(bool)), SLOT(set_use_trash(bool)));
    connect(kfoldersync_ui.auto_compare_checkBox, SIGNAL(clicked(bool)), SLOT(set_auto_compare(bool)));
    connect(kfoldersync_ui.run_in_the_background_checkBox, SIGNAL(clicked(bool)), SLOT(set_run_in_the_background(bool)));

    // Update home stats
    update_local_info();

    // A call to KXmlGuiWindow::setupGUI() populates the GUI
    // with actions, using KXMLGUI.
    // It also applies the saved mainwindow settings, if any, and ask the
    // mainwindow to automatically save settings if changed: window size,
    // toolbar position, icon size, etc.
    setupGUI();
    show();
}


KFolderSync::~KFolderSync()
{
}


void KFolderSync::setup_actions()
{
    KStandardAction::quit(KUniqueApplication::kApplication(), SLOT(closeAllWindows()), actionCollection());

    KAction* compare_action = new KAction(this);
    compare_action->setText(i18nc("@action:intoolbar", "&Compare"));
    compare_action->setIcon(KIcon("system-search"));
    actionCollection()->addAction("compare", compare_action);
    connect(compare_action, SIGNAL(triggered(bool)), SLOT(compare()));

    KAction* run_action = new KAction(this);
    run_action->setText(i18nc("@action:intoolbar", "&Run"));
    run_action->setIcon(KIcon("system-run"));
    actionCollection()->addAction("run", run_action);
    connect(run_action, SIGNAL(triggered(bool)), SLOT(run()));

    KAction* reset_app_action = new KAction(this);
    reset_app_action->setText(i18nc("@action:inmenu", "&Reset application"));
    reset_app_action->setIcon(KIcon("emblem-new"));
    actionCollection()->addAction("reset_app", reset_app_action);
    connect(reset_app_action, SIGNAL(triggered(bool)), SLOT(reset_application()));
}


void KFolderSync::update_local_info()
{
    KDiskFreeSpaceInfo info = KDiskFreeSpaceInfo::freeSpaceInfo(QDir::homePath());
    if (info.isValid()) {
        kfoldersync_ui.home_sz_bar->setMaximum(ONE_HUNDRED);
        qreal used_size = info.used();
        qreal total_size = info.size();
        kfoldersync_ui.home_sz_bar->setValue(qRound(used_size * ONE_HUNDRED / total_size));
        kfoldersync_ui.home_mount_point->setText(i18nc("@info:status", "Mount point: %1", info.mountPoint()));
        KLocale* locale = KGlobal::locale();
        kfoldersync_ui.home_free_sz->setText(i18nc("@info:status", "Available space left: %1", locale->formatByteSize(info.available())));
        kfoldersync_ui.home_used_sz->setText(i18nc("@info:status", "Used space: %1", locale->formatByteSize(used_size)));
        kfoldersync_ui.home_max_sz->setText(i18nc("@info:status", "Total size: %1", locale->formatByteSize(total_size)));
    } else {
        QString msg = i18nc("@info:status", "N/A");
        kfoldersync_ui.home_sz_bar->setMaximum(0);
        kfoldersync_ui.home_mount_point->setText(msg);
        kfoldersync_ui.home_free_sz->setText(msg);
        kfoldersync_ui.home_used_sz->setText(msg);
        kfoldersync_ui.home_max_sz->setText(msg);
    }
}


bool KFolderSync::handle_kiojob(KIO::Job* const job, const QPointer<DetailedProgressDialog>& detailed_progress_dialog)
{
    QEventLoop loop(this);

    // Create necessary connections
    if (log_requested) {
        connect(job, SIGNAL(infoMessage(KJob*,QString,QString)), logger, SLOT(log_job_info(KJob*,QString,QString)));
    }
    connect(job, SIGNAL(result(KJob*)), SLOT(check_for_errors(KJob*)));
    connect(job, SIGNAL(result(KJob*)), &loop, SLOT(quit()));
    if (detailed_progress_dialog.isNull()) {
        if (operation_must_be_canceled) {
            // return that operation must be canceled
            return true;
        }
        connect(this, SIGNAL(kill_jobs()), &loop, SLOT(quit()));
    } else {
        connect(job, SIGNAL(description(KJob*,QString,QPair<QString,QString>,QPair<QString,QString>)),
                detailed_progress_dialog, SLOT(update_job_description(KJob*,QString,QPair<QString,QString>,QPair<QString,QString>)));
        connect(job, SIGNAL(percent(KJob*,ulong)), detailed_progress_dialog, SLOT(update_progress_bars(KJob*,ulong)));
        connect(job, SIGNAL(speed(KJob*,ulong)), detailed_progress_dialog, SLOT(update_job_speed(KJob*,ulong)));
        connect(detailed_progress_dialog, SIGNAL(finished()), &loop, SLOT(quit()));
    }

    // Start loop
    loop.exec();

    // Check if operation must be canceled
    if (detailed_progress_dialog.isNull()) {
        if (operation_must_be_canceled) {
            if (job->kill(KJob::EmitResult) == false)
                loop.exec(QEventLoop::ExcludeUserInputEvents);
            // return that operation must be canceled
            return true;
        }
    } else if (detailed_progress_dialog->isHidden()) {
        operation_must_be_canceled = true;
        if (job->kill(KJob::EmitResult) == false) {
            job->disconnect(detailed_progress_dialog);
            default_progress_dialog = new KProgressDialog(this,
                    i18nc("@title:window", "Termination dialog"),
                    i18nc("@info:progress", "Canceling operations... please wait"),
                    Qt::Dialog | Qt::CustomizeWindowHint | Qt::WindowTitleHint);
            default_progress_dialog->setAllowCancel(false);
            default_progress_dialog->setModal(true);
            default_progress_dialog->setMinimumDuration(0);
            connect(job, SIGNAL(percent(KJob*,ulong)), SLOT(update_kdialog_progressBar(KJob*,ulong)));
            loop.exec(QEventLoop::ExcludeUserInputEvents);
            if (default_progress_dialog.isNull() == false)
                default_progress_dialog->delayedDestruct();
        }
        // return that operation must be canceled
        return true;
    }
    if (job_error_found) {
        if (log_requested)
            logger->log_msg(i18nc("@item:intext", "...AN ERROR OCCURRED!"));
        if (detailed_progress_dialog.isNull() || kfoldersync_ui.ignore_errors_checkBox->isChecked() == false) {
            operation_must_be_canceled = true;
            // return that operation must be canceled
            return true;
        }
    } else if (log_requested)
        logger->log_msg(i18nc("@item:intext", "...done!"));

    // Update number of completed jobs
    ++completed_jobs;

    // Return that operation must not be canceled
    return false;
}


void KFolderSync::do_sync_or_diff_backup(const KUrl& current_src_dir_kurl, const KUrl& current_dest_dir_kurl, const QPointer<DetailedProgressDialog>& detailed_progress_dialog)
{
    QList<KUrl> kurl_qlist;
    KIO::Job* job;
    KIO::UDSEntry alterego_udsentry;
    QString entry_name;
    QEventLoop loop(this);

    // Retrieve current_src_dir_kurl UDSEntryList
    KIO::ListJob* list_job = KIO::listDir(current_src_dir_kurl, KIO::HideProgressInfo);
    connect(list_job, SIGNAL(entries(KIO::Job*,KIO::UDSEntryList)), SLOT(save_udsentrylist(KIO::Job*,KIO::UDSEntryList)));
    connect(list_job, SIGNAL(result(KJob*)), &loop, SLOT(quit()));
    loop.exec();

    // Perform necessary copies (overwriting if needed)
    for (auto qlist_it = current_udsentry_list.constBegin(); qlist_it != current_udsentry_list.constEnd(); ++qlist_it) {
        entry_name = qlist_it->stringValue(KIO::UDSEntry::UDS_NAME);
        if (entry_name == "." || entry_name == "..")
            continue;
        KUrl current_kurl(current_src_dir_kurl);
        current_kurl.addPath(entry_name);
        KUrl alterego_kurl(current_dest_dir_kurl);
        alterego_kurl.addPath(entry_name);
        if (qlist_it->isDir()) {
            if (KIO::NetAccess::exists(alterego_kurl, KIO::NetAccess::DestinationSide, this)) {
                kurl_qlist.append(current_kurl);
            } else {
                job = KIO::copyAs(current_kurl, alterego_kurl, KIO::HideProgressInfo);
                if (log_requested)
                    logger->log_cp(current_kurl, alterego_kurl);
                if (handle_kiojob(job, detailed_progress_dialog))
                    // return now, because operation must be canceled
                    return;
            }
        } else {
            if (KIO::NetAccess::exists(alterego_kurl, KIO::NetAccess::DestinationSide, this)) {
                if (KIO::NetAccess::stat(alterego_kurl, alterego_udsentry, this) == false) {
                    job_error_found = true;
                    operation_must_be_canceled = true;
                    // return now, because operation must be canceled
                    return;
                }
                if (qlist_it->numberValue(KIO::UDSEntry::UDS_MODIFICATION_TIME) > alterego_udsentry.numberValue(KIO::UDSEntry::UDS_MODIFICATION_TIME)) {
                    job = KIO::file_copy(current_kurl, alterego_kurl, -1, KIO::Overwrite | KIO::HideProgressInfo);
                    if (log_requested)
                        logger->log_ovwr(alterego_kurl);
                    if (handle_kiojob(job, detailed_progress_dialog))
                        // return now, because operation must be canceled
                        return;
                }
            } else {
                job = KIO::file_copy(current_kurl, alterego_kurl, -1, KIO::HideProgressInfo);
                if (log_requested)
                    logger->log_cp(current_kurl, alterego_kurl);
                if (handle_kiojob(job, detailed_progress_dialog))
                    // return now, because operation must be canceled
                    return;
            }
        }
    }

    // Delete (or trash) useless files and directories if requested
    if (operation_type == sync) {
        // retrieve current_dest_dir_kurl UDSEntryList
        list_job = KIO::listDir(current_dest_dir_kurl, KIO::HideProgressInfo);
        connect(list_job, SIGNAL(entries(KIO::Job*,KIO::UDSEntryList)), SLOT(save_udsentrylist(KIO::Job*,KIO::UDSEntryList)));
        connect(list_job, SIGNAL(result(KJob*)), &loop, SLOT(quit()));
        loop.exec();
        // begin removals
        for (auto qlist_it = current_udsentry_list.constBegin(); qlist_it != current_udsentry_list.constEnd(); ++qlist_it) {
            entry_name = qlist_it->stringValue(KIO::UDSEntry::UDS_NAME);
            if (entry_name == "." || entry_name == "..")
                continue;
            KUrl current_kurl(current_dest_dir_kurl);
            current_kurl.addPath(entry_name);
            KUrl alterego_kurl(current_src_dir_kurl);
            alterego_kurl.addPath(entry_name);
            if (KIO::NetAccess::exists(alterego_kurl, KIO::NetAccess::DestinationSide, this) == false) {
                if (current_kurl.isLocalFile() && use_trash) {
                    job = KIO::trash(current_kurl, KIO::HideProgressInfo);
                    if (log_requested)
                        logger->log_trash(current_kurl);
                } else {
                    job = KIO::del(current_kurl, KIO::HideProgressInfo);
                    if (log_requested)
                        logger->log_del(current_kurl);
                }
                if (handle_kiojob(job, detailed_progress_dialog))
                    // return now, because operation must be canceled
                    return;
            }
        }
    }

    // Recursivity in subdirectories
    if (kurl_qlist.isEmpty() == false) {
        for (auto it = kurl_qlist.constBegin(); it != kurl_qlist.constEnd(); ++it) {
            KUrl next_dest_dir_kurl(current_dest_dir_kurl);
            next_dest_dir_kurl.addPath(it->fileName());
            do_sync_or_diff_backup(*it, next_dest_dir_kurl, detailed_progress_dialog);
            if (operation_must_be_canceled)
                return;
        }
    }

    // Finished
    return;
}


void KFolderSync::do_bilateral_sync(const KUrl& src_kurl, const KUrl& dest_kurl, const QPointer<DetailedProgressDialog>& detailed_progress_dialog)
{
    do_sync_or_diff_backup(src_kurl, dest_kurl, detailed_progress_dialog);
    if (operation_must_be_canceled)
        return;
    do_sync_or_diff_backup(dest_kurl, src_kurl, detailed_progress_dialog);

    // Finished
    return;
}


void KFolderSync::do_incr_backup(const KUrl& current_src_dir_kurl, const KUrl& current_dest_dir_kurl, const QPointer<DetailedProgressDialog>& detailed_progress_dialog, const KUrl& f_old_directory_kurl)
{
    QList<KUrl> kurl_qlist;
    KIO::Job* job;
    KIO::UDSEntry alterego_udsentry;
    QString entry_name;
    QEventLoop loop(this);

    // Retrieve current_src_dir_kurl UDSEntryList
    KIO::ListJob* list_job = KIO::listDir(current_src_dir_kurl, KIO::HideProgressInfo);
    connect(list_job, SIGNAL(entries(KIO::Job*,KIO::UDSEntryList)), SLOT(save_udsentrylist(KIO::Job*,KIO::UDSEntryList)));
    connect(list_job, SIGNAL(result(KJob*)), &loop, SLOT(quit()));
    loop.exec();

    // Perform necessary backups
    for (auto qlist_it = current_udsentry_list.constBegin(); qlist_it != current_udsentry_list.constEnd(); ++qlist_it) {
        entry_name = qlist_it->stringValue(KIO::UDSEntry::UDS_NAME);
        if (entry_name == "." || entry_name == "..")
            continue;
        KUrl current_kurl(current_src_dir_kurl);
        current_kurl.addPath(entry_name);
        KUrl alterego_kurl(current_dest_dir_kurl);
        alterego_kurl.addPath(entry_name);
        if (qlist_it->isDir()) {
            if (KIO::NetAccess::exists(alterego_kurl, KIO::NetAccess::DestinationSide, this)) {
                kurl_qlist.append(current_kurl);
            } else {
                job = KIO::copyAs(current_kurl, alterego_kurl, KIO::HideProgressInfo);
                if (log_requested)
                    logger->log_cp(current_kurl, alterego_kurl);
                if (handle_kiojob(job, detailed_progress_dialog))
                    // return now, because operation must be canceled
                    return;
            }
        } else {
            if (KIO::NetAccess::exists(alterego_kurl, KIO::NetAccess::DestinationSide, this)) {
                if (KIO::NetAccess::stat(alterego_kurl, alterego_udsentry, this) == false) {
                    job_error_found = true;
                    operation_must_be_canceled = true;
                    // return now, because operation must be canceled
                    return;
                }
                if (qlist_it->numberValue(KIO::UDSEntry::UDS_MODIFICATION_TIME) > alterego_udsentry.numberValue(KIO::UDSEntry::UDS_MODIFICATION_TIME)) {
                    KUrl tmp_kurl(f_old_directory_kurl);
                    QString alterego_new_name = alterego_kurl.fileName();
                    tmp_kurl.addPath(alterego_new_name);
                    if (KIO::NetAccess::exists(tmp_kurl, KIO::NetAccess::DestinationSide, this)) {
                        bool try_another = true;
                        quint16 version_number = 2;
                        while (try_another) {
                            KUrl test_kurl(f_old_directory_kurl);
                            alterego_new_name = alterego_kurl.fileName().append(i18nc("@item:intext", " version %1", version_number));
                            test_kurl.addPath(alterego_new_name);
                            if (KIO::NetAccess::exists(test_kurl, KIO::NetAccess::DestinationSide, this))
                                ++version_number;
                            else
                                try_another = false;
                        }
                    }
                    KUrl alterego_new_kurl(f_old_directory_kurl);
                    alterego_new_kurl.addPath(alterego_new_name);
                    job = KIO::file_move(alterego_kurl, alterego_new_kurl, -1, KIO::HideProgressInfo);
                    if (log_requested)
                        logger->log_mv(alterego_kurl, alterego_new_kurl);
                    if (handle_kiojob(job, detailed_progress_dialog))
                        // return now, because operation must be canceled
                        return;
                    job = KIO::file_copy(current_kurl, alterego_kurl, -1, KIO::HideProgressInfo);
                    if (log_requested)
                        logger->log_cp(current_kurl, alterego_kurl);
                    if (handle_kiojob(job, detailed_progress_dialog))
                        // return now, because operation must be canceled
                        return;
                }
            } else {
                job = KIO::file_copy(current_kurl, alterego_kurl, -1, KIO::HideProgressInfo);
                if (log_requested)
                    logger->log_cp(current_kurl, alterego_kurl);
                if (handle_kiojob(job, detailed_progress_dialog))
                    // return now, because operation must be canceled
                    return;
            }
        }
    }

    // Recursivity in subdirectories
    if (kurl_qlist.isEmpty() == false) {
        for (auto it = kurl_qlist.constBegin(); it != kurl_qlist.constEnd(); ++it) {
            KUrl next_dest_dir_kurl(current_dest_dir_kurl);
            next_dest_dir_kurl.addPath(it->fileName());
            do_incr_backup(*it, next_dest_dir_kurl, detailed_progress_dialog, f_old_directory_kurl);
            if (operation_must_be_canceled)
                return;
        }
    }

    // Finished
    return;
}


void KFolderSync::do_bilateral_incr_backup(const KUrl& src_kurl, const KUrl& dest_kurl, const QPointer<DetailedProgressDialog>& detailed_progress_dialog)
{
    QString f_old_directory_name = i18nc("@item", "Old files - %1", KGlobal::locale()->formatDateTime(KDateTime::currentLocalDateTime(), KLocale::LongDate, KLocale::Seconds));

    // Incremental backup from src to dest
    KUrl f_old_directory_kurl_dest(dest_kurl);
    f_old_directory_kurl_dest.addPath(f_old_directory_name);
    if (KIO::NetAccess::mkdir(f_old_directory_kurl_dest, this))
        do_incr_backup(src_kurl, dest_kurl, detailed_progress_dialog, f_old_directory_kurl_dest);
    else {
        job_error_found = true;
        operation_must_be_canceled = true;
    }
    if (operation_must_be_canceled)
        return;

    // Incremental backup from dest to src
    KUrl f_old_directory_kurl_src(src_kurl);
    f_old_directory_kurl_src.addPath(f_old_directory_name);
    if (KIO::NetAccess::mkdir(f_old_directory_kurl_src, this))
        do_incr_backup(dest_kurl, src_kurl, detailed_progress_dialog, f_old_directory_kurl_src);
    else {
        job_error_found = true;
        operation_must_be_canceled = true;
    }
    if (operation_must_be_canceled)
        return;

    // Bilateral synchronization between the two directories containing old files
    do_bilateral_sync(f_old_directory_kurl_src, f_old_directory_kurl_dest, detailed_progress_dialog);

    // Finished
    return;
}


void KFolderSync::do_full_backup(const KUrl& src_dir_kurl, KUrl dest_dir_kurl, const QPointer<DetailedProgressDialog>& detailed_progress_dialog)
{
    // Begin operation (one for each RequesterWidget)
    dest_dir_kurl.addPath(i18nc("@item",
                                "Full backup of <filename>%1</filename> - %2",
                                src_dir_kurl.fileName(),
                                KGlobal::locale()->formatDateTime(KDateTime::currentLocalDateTime(), KLocale::LongDate, KLocale::Seconds)));
    KIO::CopyJob* job = KIO::copyAs(src_dir_kurl, dest_dir_kurl, KIO::HideProgressInfo);
    if (log_requested)
        logger->log_cp(src_dir_kurl, dest_dir_kurl);
    if (handle_kiojob(job, detailed_progress_dialog))
        // return now, because operation must be canceled
        return;

    // Finished
    return;
}


void KFolderSync::analyze_udsentrylist_for_comparison(const KUrl& current_src_dir_kurl, const KUrl& current_dest_dir_kurl)
{
    KIO::DirectorySizeJob* d_sz_job;
    qint64 diff_size;
    QList<KUrl> kurl_qlist;
    KIO::UDSEntry alterego_udsentry;
    QString entry_name;
    QEventLoop loop(this);

    // Retrieve current_src_dir_kurl UDSEntryList
    KIO::ListJob* list_job = KIO::listDir(current_src_dir_kurl, KIO::HideProgressInfo);
    connect(list_job, SIGNAL(entries(KIO::Job*,KIO::UDSEntryList)), SLOT(save_udsentrylist(KIO::Job*,KIO::UDSEntryList)));
    connect(list_job, SIGNAL(result(KJob*)), &loop, SLOT(quit()));
    loop.exec(QEventLoop::ExcludeUserInputEvents);

    // Begin analysis
    for (auto qlist_it = current_udsentry_list.constBegin(); qlist_it != current_udsentry_list.constEnd(); ++qlist_it) {
        entry_name = qlist_it->stringValue(KIO::UDSEntry::UDS_NAME);
        if (entry_name == "." || entry_name == "..")
            continue;
        KUrl current_kurl(current_src_dir_kurl);
        current_kurl.addPath(entry_name);
        KUrl alterego_kurl(current_dest_dir_kurl);
        alterego_kurl.addPath(entry_name);
        if (qlist_it->isDir()) {
            if (KIO::NetAccess::exists(alterego_kurl, KIO::NetAccess::DestinationSide, this)) {
                kurl_qlist.append(current_kurl);
            } else {
                ++d_cp_n;
                d_sz_job = KIO::directorySize(current_kurl);
                connect(d_sz_job, SIGNAL(result(KJob*)), SLOT(save_dir_sz(KJob*)));
                connect(d_sz_job, SIGNAL(result(KJob*)), &loop, SLOT(quit()));
                loop.exec(QEventLoop::ExcludeUserInputEvents);
                add_sz_n += dir_size_from_job;
            }
        } else {
            if (KIO::NetAccess::exists(alterego_kurl, KIO::NetAccess::DestinationSide, this)) {
                KIO::NetAccess::stat(alterego_kurl, alterego_udsentry, this);
                if (qlist_it->numberValue(KIO::UDSEntry::UDS_MODIFICATION_TIME) > alterego_udsentry.numberValue(KIO::UDSEntry::UDS_MODIFICATION_TIME)) {
                    if (operation_type == incr_backup || operation_type == bilateral_incr_backup) {
                        ++f_old_n;
                        if (operation_type == bilateral_incr_backup)
                            f_cp_n += 2;
                        else
                            ++f_cp_n;
                        diff_size = qlist_it->numberValue(KIO::UDSEntry::UDS_SIZE);
                    } else {
                        ++f_ovwr_n;
                        diff_size = qlist_it->numberValue(KIO::UDSEntry::UDS_SIZE) - alterego_udsentry.numberValue(KIO::UDSEntry::UDS_SIZE);
                    }
                    if (diff_size < 0)
                        del_sz_n -= diff_size;
                    else
                        add_sz_n += diff_size;
                }
            } else {
                ++f_cp_n;
                add_sz_n += qlist_it->numberValue(KIO::UDSEntry::UDS_SIZE);
            }
        }
    }

    // If needed, check for necessary removals
    if (operation_type == sync) {
        // retrieve current_dest_dir_kurl UDSEntryList
        list_job = KIO::listDir(current_dest_dir_kurl, KIO::HideProgressInfo);
        connect(list_job, SIGNAL(entries(KIO::Job*,KIO::UDSEntryList)), SLOT(save_udsentrylist(KIO::Job*,KIO::UDSEntryList)));
        connect(list_job, SIGNAL(result(KJob*)), &loop, SLOT(quit()));
        loop.exec(QEventLoop::ExcludeUserInputEvents);
        // begin
        for (auto qlist_it = current_udsentry_list.constBegin(); qlist_it != current_udsentry_list.constEnd(); ++qlist_it) {
            entry_name = qlist_it->stringValue(KIO::UDSEntry::UDS_NAME);
            if (entry_name == "." || entry_name == "..")
                continue;
            KUrl current_kurl(current_dest_dir_kurl);
            current_kurl.addPath(entry_name);
            KUrl alterego_kurl(current_src_dir_kurl);
            alterego_kurl.addPath(entry_name);
            if (KIO::NetAccess::exists(alterego_kurl, KIO::NetAccess::DestinationSide, this) == false) {
                if (qlist_it->isDir()) {
                    ++d_del_n;
                    d_sz_job = KIO::directorySize(current_kurl);
                    connect(d_sz_job, SIGNAL(result(KJob*)), SLOT(save_dir_sz(KJob*)));
                    connect(d_sz_job, SIGNAL(result(KJob*)), &loop, SLOT(quit()));
                    loop.exec(QEventLoop::ExcludeUserInputEvents);
                    del_sz_n += dir_size_from_job;
                } else {
                    ++f_del_n;
                    del_sz_n += qlist_it->numberValue(KIO::UDSEntry::UDS_SIZE);
                }
            }
        }
    }

    // Recursivity in subdirectories
    if (kurl_qlist.isEmpty() == false) {
        for (auto it = kurl_qlist.constBegin(); it != kurl_qlist.constEnd(); ++it) {
            KUrl next_dest_dir_kurl(current_dest_dir_kurl);
            next_dest_dir_kurl.addPath(it->fileName());
            analyze_udsentrylist_for_comparison(*it, next_dest_dir_kurl);
        }
    }
}

void KFolderSync::do_operation()
{
    QPointer<DetailedProgressDialog> detailed_progress_dialog;
    switch (operation_type) {
    case sync:
        if (run_in_the_background == false)
            detailed_progress_dialog = new DetailedProgressDialog(completed_jobs, total_jobs, this);
        for (int k = 0; k < req_widget_list.size() && operation_must_be_canceled == false; ++k) {
            if (log_requested) {
                logger->log_start(i18nc("@item:intext", "synchronization"),
                                  req_widget_list.at(k)->source_kurl().pathOrUrl(),
                                  req_widget_list.at(k)->destination_kurl().pathOrUrl());
            }
            do_sync_or_diff_backup(req_widget_list.at(k)->source_kurl(), req_widget_list.at(k)->destination_kurl(), detailed_progress_dialog);
        }
        break;
    case bilateral_sync:
        if (run_in_the_background == false)
            detailed_progress_dialog = new DetailedProgressDialog(completed_jobs, total_jobs, this);
        for (int k = 0; k < req_widget_list.size() && operation_must_be_canceled == false; ++k) {
            if (log_requested) {
                logger->log_start(i18nc("@item:intext", "bilateral synchronization"),
                                  req_widget_list.at(k)->source_kurl().pathOrUrl(),
                                  req_widget_list.at(k)->destination_kurl().pathOrUrl());
            }
            do_bilateral_sync(req_widget_list.at(k)->source_kurl(), req_widget_list.at(k)->destination_kurl(), detailed_progress_dialog);
        }
        break;
    case diff_backup:
        if (run_in_the_background == false)
            detailed_progress_dialog = new DetailedProgressDialog(completed_jobs, total_jobs, this);
        for (int k = 0; k < req_widget_list.size() && operation_must_be_canceled == false; ++k) {
            if (log_requested) {
                logger->log_start(i18nc("@item:intext", "differential backup"),
                                  req_widget_list.at(k)->source_kurl().pathOrUrl(),
                                  req_widget_list.at(k)->destination_kurl().pathOrUrl());
            }
            do_sync_or_diff_backup(req_widget_list.at(k)->source_kurl(), req_widget_list.at(k)->destination_kurl(), detailed_progress_dialog);
        }
        break;
    case incr_backup:
        if (run_in_the_background == false)
            detailed_progress_dialog = new DetailedProgressDialog(completed_jobs, total_jobs, this);
        for (int k = 0; k < req_widget_list.size() && operation_must_be_canceled == false; ++k) {
            if (log_requested) {
                logger->log_start(i18nc("@item:intext", "incremental backup"),
                                  req_widget_list.at(k)->source_kurl().pathOrUrl(),
                                  req_widget_list.at(k)->destination_kurl().pathOrUrl());
            }
            // make the directory that will contain the old files
            KUrl f_old_directory_kurl(req_widget_list.at(k)->destination_kurl());
            f_old_directory_kurl.addPath(i18nc("@item", "Old files - %1", KGlobal::locale()->formatDateTime(KDateTime::currentLocalDateTime(), KLocale::LongDate, KLocale::Seconds)));
            if (KIO::NetAccess::mkdir(f_old_directory_kurl, this))
                do_incr_backup(req_widget_list.at(k)->source_kurl(), req_widget_list.at(k)->destination_kurl(), detailed_progress_dialog, f_old_directory_kurl);
            else
                operation_must_be_canceled = true;
        }
        break;
    case bilateral_incr_backup:
        if (run_in_the_background == false)
            detailed_progress_dialog = new DetailedProgressDialog(completed_jobs, total_jobs, this);
        for (int k = 0; k < req_widget_list.size() && operation_must_be_canceled == false; ++k) {
            if (log_requested) {
                logger->log_start(i18nc("@item:intext", "bilateral incremental backup"),
                                  req_widget_list.at(k)->source_kurl().pathOrUrl(),
                                  req_widget_list.at(k)->destination_kurl().pathOrUrl());
            }
            do_bilateral_incr_backup(req_widget_list.at(k)->source_kurl(), req_widget_list.at(k)->destination_kurl(), detailed_progress_dialog);
        }
        break;
    case full_backup:
        if (run_in_the_background == false)
            detailed_progress_dialog = new DetailedProgressDialog(completed_jobs, total_jobs, this);
        for (int k = 0; k < req_widget_list.size() && operation_must_be_canceled == false; ++k) {
            if (log_requested) {
                logger->log_start(i18nc("@item:intext", "full backup"),
                                  req_widget_list.at(k)->source_kurl().pathOrUrl(),
                                  req_widget_list.at(k)->destination_kurl().pathOrUrl());
            }
            do_full_backup(req_widget_list.at(k)->source_kurl(), req_widget_list.at(k)->destination_kurl(), detailed_progress_dialog);
        }
        break;
    default:
        qFatal("KFolderSync::do_operation() > non-existent operation type has been set");
    }
    if (run_in_the_background) {
        if (operation_must_be_canceled) {
            tray_obj.showMessage(i18nc("@title", "<application>%1</application>", KGlobal::mainComponent().aboutData()->programName()),
                                 i18nc("@info:status", "Your directories have been updated"), "task-complete", 2000);
        } else {
            tray_obj.showMessage(i18nc("@title", "<application>%1</application>", KGlobal::mainComponent().aboutData()->programName()),
                                 i18nc("@info:status", "Last operation has been interrupted"), "task-attempt", 2000);
        }
    } else if (detailed_progress_dialog.isNull() == false)
        detailed_progress_dialog->delayedDestruct();
}


void KFolderSync::set_sync_enum(const bool b)
{
    if (b) {
        operation_type = sync;
        notify_comparison_necessity();
    }
}

void KFolderSync::set_bilateral_sync_enum(const bool b)
{
    if (b) {
        operation_type = bilateral_sync;
        notify_comparison_necessity();
    }
}

void KFolderSync::set_diff_backup_enum(const bool b)
{
    if (b) {
        operation_type = diff_backup;
        notify_comparison_necessity();
    }
}

void KFolderSync::set_incr_backup_enum(const bool b)
{
    if (b) {
        operation_type = incr_backup;
        notify_comparison_necessity();
    }
}

void KFolderSync::set_bilateral_incr_backup_enum(const bool b)
{
    if (b) {
        operation_type = bilateral_incr_backup;
        notify_comparison_necessity();
    }
}

void KFolderSync::set_full_backup_enum(const bool b)
{
    if (b) {
        operation_type = full_backup;
        notify_comparison_necessity();
    }
}


void KFolderSync::init_logger(const bool b)
{
    bool succeeded;
    if (b) {
        logger = new Logger(this, succeeded);
        if (succeeded)
            log_requested = true;
        else {
            kfoldersync_ui.log_checkBox->click();
            return;
        }
    } else {
        log_requested = false;
        delete logger;
    }
}


void KFolderSync::set_use_trash(const bool b)
{
    use_trash = b;
}


void KFolderSync::set_run_in_the_background(const bool b)
{
    run_in_the_background = b;
    if (b && kfoldersync_ui.auto_compare_checkBox->isChecked() == false)
        kfoldersync_ui.auto_compare_checkBox->click();
}


void KFolderSync::set_auto_compare(const bool b)
{
    auto_compare = b;
    if (b)
        compare();
    else {
        if (kfoldersync_ui.run_in_the_background_checkBox->isChecked())
            kfoldersync_ui.run_in_the_background_checkBox->click();
    }
}


void KFolderSync::update_kdialog_progressBar(KJob*, ulong percent)
{
    default_progress_dialog->progressBar()->setValue(percent);
}


void KFolderSync::check_for_errors(KJob* job)
{
    if (job->error())
        job_error_found = true;
    else
        job_error_found = false;
}


void KFolderSync::run()
{
    // Check
    if (comparison_made)
        operation_must_be_canceled = false;
    else {
        KMessageBox::sorry(this, i18nc("@info", "<warning>you can't begin without first making a comparison!</warning>"));
        return;
    }
    if (run_in_the_background) {
        // setup background environment
        QString subtitle;
        bool go = true;
        int sec = DEFAULT_SECONDS_BETWEEN_ATTEMPTS;
        switch (operation_type) {
        case sync:
            subtitle = i18nc("@info:status", "Recurring synchronization");
            break;
        case bilateral_sync:
            subtitle = i18nc("@info:status", "Recurring bilateral synchronization");
            break;
        case diff_backup:
            subtitle = i18nc("@info:status", "Recurring differential backup");
            break;
        case incr_backup:
            subtitle = i18nc("@info:status", "Recurring incremental backup");
            sec = KInputDialog::getInteger(i18nc("@title:window", "Input dialog"), i18nc("@label", "Seconds between each backup attempt:"),
                                           DEFAULT_SECONDS_BETWEEN_ATTEMPTS, DEFAULT_SECONDS_BETWEEN_ATTEMPTS, INT_MAX / ONE_THOUSAND, 1, &go, this);
            break;
        case bilateral_incr_backup:
            subtitle = i18nc("@info:status", "Recurring bilateral incremental backup");
            sec = KInputDialog::getInteger(i18nc("@title:window", "Input dialog"), i18nc("@label", "Seconds between each backup attempt:"),
                                           DEFAULT_SECONDS_BETWEEN_ATTEMPTS, DEFAULT_SECONDS_BETWEEN_ATTEMPTS, INT_MAX / ONE_THOUSAND, 1, &go, this);
            break;
            break;
        case full_backup:
            subtitle = i18nc("@info:status", "Recurring full backup");
            sec = KInputDialog::getInteger(i18nc("@title:window", "Input dialog"), i18nc("@label", "Seconds between each backup attempt:"),
                                           DEFAULT_SECONDS_BETWEEN_ATTEMPTS, DEFAULT_SECONDS_BETWEEN_ATTEMPTS, INT_MAX / ONE_THOUSAND, 1, &go, this);
            break;
        default:
            qFatal("KFolderSync::run() > non-existent operation type has been set");
        }
        if (go == false)
            return;
        else
            msec_between_attempts = sec * ONE_THOUSAND;
        if (log_requested)
            logger->log_msg(i18nc("@item:intext", "<application>%1</application> placed into the background", KGlobal::mainComponent().aboutData()->programName()));
        tray_obj.setToolTip("task-ongoing", i18nc("@title", "<application>%1</application>", KGlobal::mainComponent().aboutData()->programName()), subtitle);
        tray_obj.setStatus(KStatusNotifierItem::Active);
        tray_obj.activate();
        connect(this, SIGNAL(kill_jobs()), SLOT(set_operation_must_be_canceled()), Qt::UniqueConnection);
        connect(&tray_obj, SIGNAL(activateRequested(bool,QPoint)), SIGNAL(kill_jobs()));
        connect(&background_timer, SIGNAL(timeout()), SLOT(trigger_background_operation()));
        background_timer.start(1);
        return;
    }
    if (auto_compare && total_jobs == 0) {
        KMessageBox::sorry(this, i18nc("@info:status", "There's nothing to do..."));
        return;
    }

    // Begin
    completed_jobs = 0;
    do_operation();

    // Update status bar
    QString current_msg;
    if (operation_must_be_canceled) {
        if (job_error_found)
            current_msg = i18nc("@info:status", "Operation has been canceled because of errors!");
        else
            current_msg = i18nc("@info:status", "Operation canceled!");
    } else {
        Q_ASSERT(completed_jobs == total_jobs);
        current_msg = i18nc("@info:status", "Mission accomplished!");
    }
    statusBar()->showMessage(current_msg);
    if (log_requested) {
        logger->log_msg(current_msg);
        logger->log_end();
    }

    // Update comparison results
    notify_comparison_necessity();
}


void KFolderSync::set_operation_must_be_canceled()
{
    operation_must_be_canceled = true;
}


void KFolderSync::trigger_background_operation()
{
    if (operation_must_be_canceled == false)
        notify_comparison_necessity();
    if (comparison_made && total_jobs > 0 && operation_must_be_canceled == false) {
        KIO::DirectorySizeJob* d_sz_job;
        QList<KIO::filesize_t> src_sizes_list;
        bool srcs_busy;
        quint8 checks = 0;
        int msec_to_wait;
        QEventLoop loop(this);
        for (int k = 0; k < req_widget_list.size() && operation_must_be_canceled == false; ++k) {
            d_sz_job = KIO::directorySize(req_widget_list.at(k)->source_kurl());
            connect(d_sz_job, SIGNAL(result(KJob*)), SLOT(save_dir_sz(KJob*)));
            connect(d_sz_job, SIGNAL(result(KJob*)), &loop, SLOT(quit()));
            loop.exec(QEventLoop::ExcludeUserInputEvents);
            const KIO::filesize_t new_size = dir_size_from_job;
            src_sizes_list.append(new_size);

        }
        while (operation_must_be_canceled == false && checks < 3) {
            srcs_busy = false;
            for (int k = 0; k < req_widget_list.size() && operation_must_be_canceled == false; ++k) {
                d_sz_job = KIO::directorySize(req_widget_list.at(k)->source_kurl());
                connect(d_sz_job, SIGNAL(result(KJob*)), SLOT(save_dir_sz(KJob*)));
                connect(d_sz_job, SIGNAL(result(KJob*)), &loop, SLOT(quit()));
                loop.exec(QEventLoop::ExcludeUserInputEvents);
                if (src_sizes_list.at(k) != dir_size_from_job) {
                    srcs_busy = true;
                    const KIO::filesize_t new_size = dir_size_from_job;
                    src_sizes_list.replace(k, new_size);
                }
            }
            if (srcs_busy) {
                checks = 0;
                msec_to_wait = msec_between_attempts * 2;
            } else {
                ++checks;
                msec_to_wait = 6000;
            }
            if (operation_must_be_canceled == false) {
                QTimer::singleShot(msec_to_wait, &loop, SLOT(quit()));
                connect(this, SIGNAL(kill_jobs()), &loop, SLOT(quit()));
                loop.exec(QEventLoop::ExcludeUserInputEvents);
            }
        }
        if (operation_must_be_canceled == false)
            do_operation();
    }
    if (operation_must_be_canceled)
        QTimer::singleShot(1, this, SLOT(come_back_into_the_foreground()));
    else
        background_timer.start(msec_between_attempts);
}


void KFolderSync::come_back_into_the_foreground()
{
    tray_obj.disconnect();
    tray_obj.setStatus(KStatusNotifierItem::Passive);
    tray_obj.setToolTip("kfoldersync", QString(), QString());
    background_timer.disconnect();

    // Update comparison results after returning
    notify_comparison_necessity();

    // Update status bar
    statusBar()->showMessage(i18nc("@info:status", "Background operations have been stopped!"));
    if (log_requested) {
        logger->log_msg(i18nc("@item:intext", "<application>%1</application> placed back into the foreground", KGlobal::mainComponent().aboutData()->programName()));
        logger->log_end();
    }
}


void KFolderSync::save_udsentrylist(KIO::Job*, KIO::UDSEntryList tmp_udsentry_list)
{
    current_udsentry_list = tmp_udsentry_list;
}


void KFolderSync::save_dir_sz(KJob* job)
{
    check_for_errors(job);
    KIO::DirectorySizeJob* current_job = qobject_cast<KIO::DirectorySizeJob*>(job);
    dir_size_from_job = current_job->totalSize();
}


void KFolderSync::notify_comparison_necessity()
{
    if (auto_compare)
        compare();
    else {
        comparison_made = false;
        statusBar()->showMessage(i18nc("@info:status", "Comparison results are out-of-date"));
    }
}


void KFolderSync::compare()
{
    if (isVisible()) {
        statusBar()->clearMessage();
        // check if all requesterwidgets are ready
        for (int k = 0; k < req_widget_list.size(); ++k) {
            if (req_widget_list.at(k)->isReady() == false) {
                // not ready to compare
                if (auto_compare == false)
                    KMessageBox::sorry(this, i18nc("@info", "<warning>you must first select all the directories!</warning>"));
                return;
            }
        }
        QList<KUrl> tmp_dest_dirs_list;
        if (operation_type == sync) {
            // check for duplicates in destination kurls
            for (int k = 0; k < req_widget_list.size(); ++k) {
                if (tmp_dest_dirs_list.contains(req_widget_list.at(k)->destination_kurl())) {
                    KMessageBox::sorry(this, i18nc("@info",
                                                   "<warning>2 identical destinations directories detected!</warning><nl/><nl/>"
                                                   "<note>performing multiple synchronizations with the same destination directory<nl/>"
                                                   "isn't very reasonable, because, obviously, only the last one will show its effects.<nl/>"
                                                   "That said, you can choose another operation type or another destination directory.</note>"));
                    req_widget_list.at(k)->clear_dest();
                    return;
                } else
                    tmp_dest_dirs_list.append(req_widget_list.at(k)->destination_kurl());
            }
        }
        // setup progress dialog
        default_progress_dialog = new KProgressDialog(this,
                i18nc("@title:window", "Comparison dialog"),
                i18nc("@info:progress", "Comparing... please wait.<nl/><nl/>"
                      "<note>network folders may require several time.</note>"),
                Qt::Dialog | Qt::CustomizeWindowHint | Qt::WindowTitleHint);
        default_progress_dialog->setAllowCancel(false);
        default_progress_dialog->setModal(true);
        default_progress_dialog->progressBar()->setMaximum(0);
        default_progress_dialog->setMinimumDuration(0);
    }

    // Initialize variables
    f_cp_n = 0;
    f_ovwr_n = 0;
    f_old_n = 0;
    f_del_n = 0;
    d_cp_n = 0;
    d_del_n = 0;
    add_sz_n = 0;
    del_sz_n = 0;
    net_ch_sz_n = 0;

    // Begin operation (one for each RequesterWidget)
    if (operation_type == full_backup) {
        d_cp_n = req_widget_list.size();
        QEventLoop loop(this);
        for (uint k = 0; k < d_cp_n; ++k) {
            KIO::DirectorySizeJob* d_sz_job = KIO::directorySize(req_widget_list.at(k)->source_kurl());
            connect(d_sz_job, SIGNAL(result(KJob*)), SLOT(save_dir_sz(KJob*)));
            connect(d_sz_job, SIGNAL(result(KJob*)), &loop, SLOT(quit()));
            loop.exec(QEventLoop::ExcludeUserInputEvents);
            add_sz_n += dir_size_from_job;
        }
    } else
        for (int k = 0; k < req_widget_list.size(); ++k) {
            analyze_udsentrylist_for_comparison(req_widget_list.at(k)->source_kurl(), req_widget_list.at(k)->destination_kurl());
            if (operation_type == bilateral_sync || operation_type == bilateral_incr_backup)
                analyze_udsentrylist_for_comparison(req_widget_list.at(k)->destination_kurl(), req_widget_list.at(k)->source_kurl());
        }

    // Update data
    total_jobs = f_cp_n + f_ovwr_n + f_old_n + f_del_n + d_cp_n + d_del_n;
    net_ch_sz_n = add_sz_n - del_sz_n;
    comparison_made = true;

    // Refresh GUI
    if (isVisible()) {
        KLocale* const locale = KGlobal::locale();
        kfoldersync_ui.f_cp_num_label->setText(locale->formatNumber(f_cp_n, 0));
        kfoldersync_ui.f_ovwr_num_label->setText(locale->formatNumber(f_ovwr_n, 0));
        kfoldersync_ui.f_old_num_label->setText(locale->formatNumber(f_old_n, 0));
        kfoldersync_ui.f_del_num_label->setText(locale->formatNumber(f_del_n, 0));
        kfoldersync_ui.d_cp_num_label->setText(locale->formatNumber(d_cp_n, 0));
        kfoldersync_ui.d_del_num_label->setText(locale->formatNumber(d_del_n, 0));
        kfoldersync_ui.add_sz_num_label->setText(locale->formatByteSize(add_sz_n));
        kfoldersync_ui.del_sz_num_label->setText(locale->formatByteSize(del_sz_n));
        kfoldersync_ui.net_ch_sz_num_label->setText(locale->formatByteSize(net_ch_sz_n));
        if (total_jobs > 0) {
            kfoldersync_ui.f_cp_progressBar->setValue(qRound(f_cp_n * ONE_HUNDRED / total_jobs));
            kfoldersync_ui.f_ovwr_progressBar->setValue(qRound(f_ovwr_n * ONE_HUNDRED / total_jobs));
            kfoldersync_ui.f_old_progressBar->setValue(qRound(f_old_n * ONE_HUNDRED / total_jobs));
            kfoldersync_ui.f_del_progressBar->setValue(qRound(f_del_n * ONE_HUNDRED / total_jobs));
            kfoldersync_ui.d_cp_progressBar->setValue(qRound(d_cp_n * ONE_HUNDRED / total_jobs));
            kfoldersync_ui.d_del_progressBar->setValue(qRound(d_del_n * ONE_HUNDRED / total_jobs));
        } else {
            kfoldersync_ui.f_cp_progressBar->reset();
            kfoldersync_ui.f_ovwr_progressBar->reset();
            kfoldersync_ui.f_old_progressBar->reset();
            kfoldersync_ui.f_del_progressBar->reset();
            kfoldersync_ui.d_cp_progressBar->reset();
            kfoldersync_ui.d_del_progressBar->reset();
        }
        update_local_info();
        if (total_jobs > 0) {
            statusBar()->showMessage(i18nc("@info:status", "Ready to begin"));
        }
        if (default_progress_dialog.isNull() == false)
            default_progress_dialog->delayedDestruct();
    }
}


void KFolderSync::add_req_widget()
{
    // Initialize widget
    RequesterWidget* tmp_w = new RequesterWidget(this, req_widget_list.size());
    kfoldersync_ui.req_w_vLayout->addWidget(tmp_w);
    connect(tmp_w, SIGNAL(destroyed(int)), SLOT(del_req_widget(int)));
    connect(tmp_w, SIGNAL(ready_to_go()), SLOT(notify_comparison_necessity()));
    req_widget_list.append(tmp_w);

    // Handle GUI consistency
    if (req_widget_list.size() > 1) {
        for (int k = 0; k < req_widget_list.size(); ++k)
            req_widget_list.at(k)->set_button_enabled(true);
        kfoldersync_ui.label_src->setText(i18nc("@title:column", "Source directories"));
        kfoldersync_ui.label_dest->setText(i18nc("@title:column", "Destination directories"));
    }
}


void KFolderSync::del_req_widget(const int current_index_pos)
{
    // Remove requester widget
    req_widget_list.removeAt(current_index_pos);

    // Recalculate requester widgets list indexes
    for (int k = 0; k < req_widget_list.size(); ++k)
        req_widget_list.at(k)->set_index_pos(k);

    // Handle GUI consistency
    if (req_widget_list.size() < 2) {
        req_widget_list.at(0)->set_button_enabled(false);
        kfoldersync_ui.label_src->setText(i18nc("@title:column", "Source directory"));
        kfoldersync_ui.label_dest->setText(i18nc("@title:column", "Destination directory"));
    }
    notify_comparison_necessity();
}


void KFolderSync::reset_application()
{
    if (KMessageBox::warningContinueCancel(this, i18nc("@info",
                                           "<warning><application>%1</application> is going to be reset.</warning> Do you really want to proceed?",
                                           KGlobal::mainComponent().aboutData()->programName())) != KMessageBox::Continue) {
        return;
    }
    hide();
    QFile::remove(localXMLFile());
    QFile::remove(KGlobal::dirs()->locateLocal("config", KGlobal::config()->name()));
    KGlobal::config()->reparseConfiguration();
    new KFolderSync;
    deleteLater();
}
