/********************************************************************************
 * Copyright (C) 2014 by Lorenzo Porta (Vindex17) <vindex17@outlook.it>         *
 *                                                                              *
 * This program is free software; you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as               *
 * published by the Free Software Foundation; either version 2 of               *
 * the License or (at your option) version 3 or any later version               *
 * accepted by the membership of KDE e.V. (or its successor approved            *
 * by the membership of KDE e.V.), which shall act as a proxy                   *
 * defined in Section 14 of version 3 of the license.                           *
 *                                                                              *
 * This program is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of               *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
 * GNU General Public License for more details.                                 *
 *                                                                              *
 * You should have received a copy of the GNU General Public License            *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                              *
 ********************************************************************************/

#include "requesterwidget.h"
#include "requesterwidget.moc"

#include <KLineEdit>
#include <KMessageBox>
#include <QFileInfo>
#include <QTimer>

RequesterWidget::RequesterWidget(QWidget* parent, int i): QWidget(parent)
{
    index_pos = i;
    requesterwidget_ui.setupUi(this);
    KUrl start_dir(QDir::homePath());

    requesterwidget_ui.src_kurlrequester->lineEdit()->setReadOnly(true);
    requesterwidget_ui.src_kurlrequester->setStartDir(start_dir);
    connect(requesterwidget_ui.src_kurlrequester, SIGNAL(urlSelected(KUrl)), SLOT(set_src_dir_kurl(KUrl)));

    requesterwidget_ui.dest_kurlrequester->lineEdit()->setReadOnly(true);
    requesterwidget_ui.dest_kurlrequester->setStartDir(start_dir);
    connect(requesterwidget_ui.dest_kurlrequester, SIGNAL(urlSelected(KUrl)), SLOT(set_dest_dir_kurl(KUrl)));

    connect(requesterwidget_ui.kpushbutton, SIGNAL(pressed()), SLOT(destroy()));
}


bool RequesterWidget::isReady()
{
    return ready;
}


void RequesterWidget::set_index_pos(int i)
{
    index_pos = i;
}


void RequesterWidget::set_button_enabled(bool b)
{
    requesterwidget_ui.kpushbutton->setEnabled(b);
}


const KUrl RequesterWidget::source_kurl()
{
    return src_dir_kurl;
}


const KUrl RequesterWidget::destination_kurl()
{
    return dest_dir_kurl;
}


const QString RequesterWidget::sameURL_error_msg()
{
    return i18nc("@info", "<warning>a folder can't be the source and the destination at the same time!</warning>");
}


const QString RequesterWidget::relationship_error_msg(const KUrl& parent_kurl, const KUrl& child_kurl)
{
    return i18nc("@info",
                 "<warning><filename>%1</filename> is parent of <filename>%2</filename>!!</warning>"
                 "<nl/><nl/><note>operations between related directories may wage war on your PC...</note>",
                 parent_kurl.pathOrUrl(),
                 child_kurl.pathOrUrl()
                );
}


bool RequesterWidget::are_related(const KUrl& current_url, const KUrl& other_url)
{
    if (current_url.isParentOf(other_url)) {
        KMessageBox::sorry(this, relationship_error_msg(current_url, other_url));
        return true;
    }
    if (other_url.isParentOf(current_url)) {
        KMessageBox::sorry(this, relationship_error_msg(other_url, current_url));
        return true;
    }

    return false;
}


void RequesterWidget::clear_src()
{
    requesterwidget_ui.src_kurlrequester->clear();
    src_dir_kurl.clear();
    ready = false;
}


void RequesterWidget::clear_dest()
{
    requesterwidget_ui.dest_kurlrequester->clear();
    dest_dir_kurl.clear();
    ready = false;
}


void RequesterWidget::set_src_dir_kurl(KUrl current_url)
{
    if (current_url.isLocalFile()) {
        if (QFileInfo(current_url.path()).isReadable() == false) {
            QString message = i18nc("@info", "<warning><filename>%1</filename> hasn't read permissions!</warning>",
                                    current_url.path(KUrl::AddTrailingSlash));
            KMessageBox::sorry(this, message);
            clear_src();
            return;
        }
    }
    if (dest_dir_kurl.isEmpty() == false) {
        if (current_url == dest_dir_kurl) {
            KMessageBox::sorry(this, sameURL_error_msg());
            clear_src();
            return;
        }
        if (are_related(current_url, dest_dir_kurl)) {
            clear_src();
            return;
        }
        ready = true;
    }
    src_dir_kurl = current_url;
    if (ready)
        emit ready_to_go();
}


void RequesterWidget::set_dest_dir_kurl(KUrl current_url)
{
    if (current_url.isLocalFile()) {
        if (QFileInfo(current_url.path()).isWritable() == false) {
            QString message = i18nc("@info", "<warning><filename>%1</filename> hasn't write permissions!</warning>",
                                    current_url.path(KUrl::AddTrailingSlash));
            KMessageBox::sorry(this, message);
            clear_dest();
            return;
        }
    }
    if (src_dir_kurl.isEmpty() == false) {
        if (current_url == src_dir_kurl) {
            KMessageBox::sorry(this, sameURL_error_msg());
            clear_dest();
            return;
        }
        if (are_related(current_url, src_dir_kurl)) {
            clear_dest();
            return;
        }
        ready = true;
    }
    dest_dir_kurl = current_url;
    if (ready)
        emit ready_to_go();
}


void RequesterWidget::destroy()
{
    hide();
    emit destroyed(index_pos);
    deleteLater();
}
